@@ -1,5 +1,5 @@
 /* Miscellaneous global declarations and portability cruft for GNU Make.
-Copyright (C) 1988-2022 Free Software Foundation, Inc.
+Copyright (C) 1988-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -12,34 +12,12 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* We use <config.h> instead of "config.h" so that a compilation
    using -I. -I$srcdir will use ./config.h rather than $srcdir/config.h
    (which it would do because makeint.h was found in $srcdir).  */
 #include <config.h>
-#undef  HAVE_CONFIG_H
-#define HAVE_CONFIG_H 1
-
-/* Specify we want GNU source code.  This must be defined before any
-   system headers are included.  */
-
-#define _GNU_SOURCE 1
-
-/* AIX requires this to be the first thing in the file.  */
-#if HAVE_ALLOCA_H
-# include <alloca.h>
-#else
-# ifdef _AIX
- #pragma alloca
-# else
-#  if !defined(__GNUC__) && !defined(WINDOWS32)
-#   ifndef alloca /* predefined by HP cc +Olibcalls */
-char *alloca ();
-#   endif
-#  endif
-# endif
-#endif
 
 /* Some versions of GCC (e.g., 10.x) set the warn_unused_result attribute on
    __builtin_alloca.  This causes alloca(0) to fail and is not easily worked
@@ -82,7 +60,6 @@ char *alloca ();
 # define __NO_STRING_INLINES
 #endif
 
-#include <stddef.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <signal.h>
@@ -105,6 +82,11 @@ char *alloca ();
 extern int errno;
 #endif
 
+/* Define macros specifying which OS we are building for.  */
+#if defined(__MVS__)
+# define MK_OS_ZOS 1
+#endif
+
 #ifdef __VMS
 /* In strict ANSI mode, VMS compilers should not be defining the
    VMS macro.  Define it here instead of a bulk edit for the correct code.
@@ -152,13 +134,14 @@ extern int errno;
 #endif
 
 #ifndef PATH_MAX
-# ifndef POSIX
+# ifdef MAXPATHLEN
 #  define PATH_MAX      MAXPATHLEN
+# else
+/* Some systems (HURD) have fully dynamic pathnames with no maximum.
+   Ideally we'd support this but it will take some work.  */
+#  define PATH_MAX      4096
 # endif
 #endif
-#ifndef MAXPATHLEN
-# define MAXPATHLEN 1024
-#endif
 
 #ifdef  PATH_MAX
 # define GET_PATH_MAX   PATH_MAX
@@ -250,8 +233,6 @@ extern int vms_unix_simulation;
 # ifdef HAVE_STRING_H
 #  include <string.h>
 #  define ANSI_STRING 1
-# else
-#  include <strings.h>
 # endif
 # ifdef HAVE_MEMORY_H
 #  include <memory.h>
@@ -294,6 +275,25 @@ char *strerror (int errnum);
 #if HAVE_STDINT_H
 # include <stdint.h>
 #endif
+
+#if HAVE_STRINGS_H
+# include <strings.h>  /* Needed for strcasecmp / strncasecmp.  */
+#endif
+
+#if defined _MSC_VER || defined __MINGW32__
+# define MK_PRI64_PREFIX "I64"
+#else
+# define MK_PRI64_PREFIX "ll"
+#endif
+#ifndef PRIdMAX
+# define PRIdMAX MK_PRI64_PREFIX "d"
+#endif
+#ifndef PRIuMAX
+# define PRIuMAX MK_PRI64_PREFIX "u"
+#endif
+#ifndef SCNdMAX
+# define SCNdMAX PRIdMAX
+#endif
 #define FILE_TIMESTAMP uintmax_t
 
 #if !defined(HAVE_STRSIGNAL)
@@ -398,6 +398,7 @@ extern int unixy_shell;
 #define NONE_SET(_v,_m) (! ANY_SET ((_v),(_m)))
 #define ALL_SET(_v,_m)  (((_v)&(_m)) == (_m))
 
+/* Bitmasks for the STOPCHAR array.  */
 #define MAP_NUL         0x0001
 #define MAP_BLANK       0x0002  /* space, TAB */
 #define MAP_NEWLINE     0x0004
@@ -447,6 +448,7 @@ extern int unixy_shell;
 # define MAP_PATHSEP     MAP_SEMI
 #elif PATH_SEPARATOR_CHAR == ','
 # define MAP_PATHSEP     MAP_COMMA
+
 #else
 # error "Unknown PATH_SEPARATOR_CHAR"
 #endif
@@ -462,6 +464,18 @@ extern int unixy_shell;
 /* Move S past all whitespace (including newlines).  */
 #define NEXT_TOKEN(s)   while (ISSPACE (*(s))) ++(s)
 
+/* True if C is a directory separator on the current system.  */
+#define ISDIRSEP(c)     STOP_SET((c),MAP_DIRSEP)
+
+/* True if S starts with a drive specifier.  */
+#if defined(HAVE_DOS_PATHS)
+# define HAS_DRIVESPEC(_s) ((((_s)[0] >= 'a' && (_s)[0] <= 'z')          \
+                             || ((_s)[0] >= 'A' && (_s)[0] <= 'Z'))      \
+                            && (_s)[1] == ':')
+#else
+# define HAS_DRIVESPEC(_s) 0
+#endif
+
 /* We can't run setrlimit when using posix_spawn.  */
 #if defined(HAVE_SYS_RESOURCE_H) && defined(HAVE_GETRLIMIT) && defined(HAVE_SETRLIMIT) && !defined(USE_POSIX_SPAWN)
 # define SET_STACK_SIZE
@@ -477,6 +491,8 @@ extern struct rlimit stack_limit;
 
 /* Number of characters in a string constant.  Does NOT include the \0 byte.  */
 #define CSTRLEN(_s)           (sizeof (_s)-1)
+
+/* Only usable when NOT calling a macro: only use it for local functions.  */
 #define STRING_SIZE_TUPLE(_s) (_s), CSTRLEN(_s)
 
 /* The number of bytes needed to represent the largest signed and unsigned
@@ -492,6 +508,15 @@ extern struct rlimit stack_limit;
 # define TTYNAME(_f) DEFAULT_TTYNAME
 #endif
 
+#ifdef VMS
+# define DEFAULT_TMPDIR     "/sys$scratch/"
+#elif defined(P_tmpdir)
+# define DEFAULT_TMPDIR     P_tmpdir
+#else
+# define DEFAULT_TMPDIR     "/tmp"
+#endif
+
+
 
 
 struct file;
@@ -511,7 +536,7 @@ void error (const floc *flocp, size_t length, const char *fmt, ...)
             ATTRIBUTE ((__format__ (__printf__, 3, 4)));
 void fatal (const floc *flocp, size_t length, const char *fmt, ...)
             ATTRIBUTE ((noreturn, __format__ (__printf__, 3, 4)));
-void out_of_memory () NORETURN;
+void out_of_memory (void) NORETURN;
 
 /* When adding macros to this list be sure to update the value of
    XGETTEXT_OPTIONS in the po/Makevars file.  */
@@ -529,13 +554,23 @@ void out_of_memory () NORETURN;
 #define ONS(_t,_a,_f,_n,_s)   _t((_a), INTSTR_LENGTH + strlen (_s), \
                                  (_f), (_n), (_s))
 
-void decode_env_switches (const char*, size_t line);
+enum variable_origin;
+struct variable;
+
+void reset_makeflags (enum variable_origin origin);
+struct variable *define_makeflags (int makefile);
+int should_print_dir (void);
+void temp_stdin_unlink (void);
 void die (int) NORETURN;
 void pfatal_with_name (const char *) NORETURN;
 void perror_with_name (const char *, const char *);
 #define xstrlen(_s) ((_s)==NULL ? 0 : strlen (_s))
 unsigned int make_toui (const char*, const char**);
-pid_t make_pid ();
+char *make_lltoa (long long, char *);
+char *make_ulltoa (unsigned long long, char *);
+void make_seed (unsigned int);
+unsigned int make_rand (void);
+pid_t make_pid (void);
 void *xmalloc (size_t);
 void *xcalloc (size_t);
 void *xrealloc (void *, size_t);
@@ -550,7 +585,7 @@ int alpha_compare (const void *, const void *);
 void print_spaces (unsigned int);
 char *find_percent (char *);
 const char *find_percent_cached (const char **);
-char *get_tmppath ();
+const char *get_tmpdir (void);
 int get_tmpfd (char **);
 FILE *get_tmpfile (char **);
 ssize_t writebuf (int, const void *, size_t);
@@ -566,13 +601,14 @@ void ar_parse_name (const char *, char **, char **);
 int ar_touch (const char *);
 time_t ar_member_date (const char *);
 
-typedef long int (*ar_member_func_t) (int desc, const char *mem, int truncated,
+typedef intmax_t (*ar_member_func_t) (int desc, const char *mem, int truncated,
                                       long int hdrpos, long int datapos,
-                                      long int size, long int date, int uid,
+                                      long int size, intmax_t date, int uid,
                                       int gid, unsigned int mode,
                                       const void *arg);
 
-long int ar_scan (const char *archive, ar_member_func_t function, const void *arg);
+intmax_t ar_scan (const char *archive, ar_member_func_t function,
+                  const void *arg);
 int ar_name_equal (const char *name, const char *mem, int truncated);
 #ifndef VMS
 int ar_member_touch (const char *arname, const char *memname);
@@ -625,8 +661,12 @@ int unload_file (const char *name);
 #ifdef MAKE_MAINTAINER_MODE
 # define SPIN(_s) spin (_s)
 void spin (const char* suffix);
+# define DBG(_f) dbg _f
+void dbg (const char *fmt, ...);
 #else
 # define SPIN(_s)
+/* Never put this code into Git or a release.  */
+# define DBG(_f) compile-error
 #endif
 
 /* We omit these declarations on non-POSIX systems which define _POSIX_VERSION,
@@ -640,10 +680,10 @@ long int lseek ();
 
 # ifdef  HAVE_GETCWD
 #  if !defined(VMS) && !defined(__DECC)
-char *getcwd ();
+char *getcwd (void);
 #  endif
 # else
-char *getwd ();
+char *getwd (void);
 #  define getcwd(buf, len)       getwd (buf)
 # endif
 
@@ -654,9 +694,6 @@ char *getwd ();
 #  define strcasecmp stricmp
 # elif HAVE_STRCMPI
 #  define strcasecmp strcmpi
-# else
-/* Create our own, in misc.c */
-int strcasecmp (const char *s1, const char *s2);
 # endif
 #endif
 
@@ -665,22 +702,9 @@ int strcasecmp (const char *s1, const char *s2);
 #  define strncasecmp strnicmp
 # elif HAVE_STRNCMPI
 #  define strncasecmp strncmpi
-# else
-/* Create our own, in misc.c */
-int strncasecmp (const char *s1, const char *s2, int n);
 # endif
 #endif
 
-#if !HAVE_MEMPCPY
-/* Create our own, in misc.c */
-void *mempcpy (void *dest, const void *src, size_t n);
-#endif
-
-#if !HAVE_STPCPY
-/* Create our own, in misc.c */
-char *stpcpy (char *dest, const char *src);
-#endif
-
 #define OUTPUT_SYNC_NONE    0
 #define OUTPUT_SYNC_LINE    1
 #define OUTPUT_SYNC_TARGET  2
@@ -697,7 +721,7 @@ extern unsigned short stopchar_map[];
 extern int just_print_flag, run_silent, ignore_errors_flag, keep_going_flag;
 extern int print_data_base_flag, question_flag, touch_flag, always_make_flag;
 extern int env_overrides, no_builtin_rules_flag, no_builtin_variables_flag;
-extern int print_version_flag, print_directory, check_symlink_flag;
+extern int print_version_flag, check_symlink_flag;
 extern int warn_undefined_variables_flag, posix_pedantic;
 extern int not_parallel, second_expansion, clock_skew_detected;
 extern int rebuilding_makefiles, one_shell, output_sync, verify_flag;
@@ -716,6 +740,8 @@ extern int batch_mode_shell;
 #define RECIPEPREFIX_DEFAULT    '\t'
 extern char cmd_prefix;
 
+extern unsigned int no_intermediates;
+
 #define JOBSERVER_AUTH_OPT      "jobserver-auth"
 
 extern char *jobserver_auth;
@@ -771,7 +797,7 @@ extern char *version_string, *remote_description, *make_host;
 
 extern unsigned int commands_started;
 
-extern int handling_fatal_signal;
+extern volatile sig_atomic_t handling_fatal_signal;
 
 #ifndef MIN
 #define MIN(_a,_b) ((_a)<(_b)?(_a):(_b))
