@@ -1,5 +1,5 @@
 /* Basic dependency engine for GNU Make.
-Copyright (C) 1988-2022 Free Software Foundation, Inc.
+Copyright (C) 1988-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -34,10 +34,11 @@ this program.  If not, see <https://www.gnu.org/licenses/>.  */
 #include <starlet.h>
 #endif
 #ifdef WINDOWS32
+#include <windows.h>
 #include <io.h>
 #include <sys/stat.h>
 #if defined(_MSC_VER) && _MSC_VER > 1200
-/* VC7 or later supprots _stat64 to access 64-bit file size. */
+/* VC7 or later supports _stat64 to access 64-bit file size. */
 #define STAT _stat64
 #else
 #define STAT stat
@@ -81,19 +82,20 @@ static const char *library_search (const char *lib, FILE_TIMESTAMP *mtime_ptr);
 static void
 check_also_make (const struct file *file)
 {
-  /* If the target was created by an implicit rule, and it exists and was
-     updated, warn about any of its also_make targets that don't exist.  */
-  if (file->tried_implicit && is_ordinary_mtime (file->last_mtime)
-      && file->last_mtime > file->mtime_before_update)
-    {
-      struct dep *ad;
+  struct dep *ad;
+  FILE_TIMESTAMP mtime = file->last_mtime;
 
-      for (ad = file->also_make; ad; ad = ad->next)
-        if (ad->file->last_mtime == NONEXISTENT_MTIME)
-          OS (error, file->cmds ? &file->cmds->fileinfo : NILF,
-              _("warning: pattern recipe did not update peer target '%s'."),
-              ad->file->name);
-    }
+  if (mtime == UNKNOWN_MTIME)
+    mtime = name_mtime (file->name);
+
+  /* If we updated the file, check its also-make files.  */
+
+  if (is_ordinary_mtime (mtime) && mtime > file->mtime_before_update)
+    for (ad = file->also_make; ad; ad = ad->next)
+      if (ad->file->last_mtime == NONEXISTENT_MTIME)
+        OS (error, file->cmds ? &file->cmds->fileinfo : NILF,
+            _("warning: pattern recipe did not update peer target '%s'."),
+            ad->file->name);
 }
 
 /* Remake all the goals in the 'struct dep' chain GOALS.  Return update_status
@@ -205,8 +207,6 @@ update_goal_chain (struct goaldep *goaldeps)
                       FILE_TIMESTAMP mtime = MTIME (file);
                       check_renamed (file);
 
-                      check_also_make (file);
-
                       if (file->updated && mtime != file->mtime_before_update)
                         {
                           /* Updating was done.  If this is a makefile and
@@ -871,7 +871,8 @@ update_file_1 (struct file *file, unsigned int depth)
 
       /* Since make has not created this file, make should not remove it,
          even if the file is intermediate. */
-      file->secondary = 1;
+      if (!file->notintermediate && no_intermediates == 0)
+        file->secondary = 1;
 
       notice_finished_file (file);
 
@@ -1039,23 +1040,30 @@ notice_finished_file (struct file *file)
     }
 
   if (ran && file->update_status != us_none)
-    /* We actually tried to update FILE, which has
-       updated its also_make's as well (if it worked).
-       If it didn't work, it wouldn't work again for them.
-       So mark them as updated with the same status.  */
-    for (d = file->also_make; d != 0; d = d->next)
-      {
-        d->file->command_state = cs_finished;
-        d->file->updated = 1;
-        d->file->update_status = file->update_status;
+    {
+      /* We actually tried to update FILE, which has
+         updated its also_make's as well (if it worked).
+         If it didn't work, it wouldn't work again for them.
+         So mark them as updated with the same status.  */
+      for (d = file->also_make; d != 0; d = d->next)
+        {
+          d->file->command_state = cs_finished;
+          d->file->updated = 1;
+          d->file->update_status = file->update_status;
 
-        if (ran && !d->file->phony)
-          /* Fetch the new modification time.
-             We do this instead of just invalidating the cached time
-             so that a vpath_search can happen.  Otherwise, it would
-             never be done because the target is already updated.  */
-          f_mtime (d->file, 0);
-      }
+          if (ran && !d->file->phony)
+            /* Fetch the new modification time.
+               We do this instead of just invalidating the cached time
+               so that a vpath_search can happen.  Otherwise, it would
+               never be done because the target is already updated.  */
+            f_mtime (d->file, 0);
+        }
+
+      /* If the target was created by an implicit rule, and it was updated,
+         warn about any of its also_make targets that don't exist.  */
+      if (file->tried_implicit && file->also_make)
+        check_also_make (file);
+    }
   else if (file->update_status == us_none)
     /* Nothing was done for FILE, but it needed nothing done.
        So mark it now as "succeeded".  */
@@ -1094,7 +1102,6 @@ check_dep (struct file *file, unsigned int depth,
       check_renamed (file);
       if (mtime == NONEXISTENT_MTIME || mtime > this_mtime)
         *must_make_ptr = 1;
-      check_also_make (file);
     }
   else
     {
@@ -1334,7 +1341,7 @@ f_mtime (struct file *file, int search)
   if (ar_name (file->name))
     {
       /* This file is an archive-member reference.  */
-
+      FILE_TIMESTAMP memmtime;
       char *arname, *memname;
       struct file *arfile;
       time_t member_date;
@@ -1342,6 +1349,9 @@ f_mtime (struct file *file, int search)
       /* Find the archive's name.  */
       ar_parse_name (file->name, &arname, &memname);
 
+      /* Find the mtime of the member file (it might not exist).  */
+      memmtime = name_mtime (memname);
+
       /* Find the modification time of the archive itself.
          Also allow for its name to be changed via VPATH search.  */
       arfile = lookup_file (arname);
@@ -1385,9 +1395,16 @@ f_mtime (struct file *file, int search)
         return NONEXISTENT_MTIME;
 
       member_date = ar_member_date (file->hname);
-      mtime = (member_date == (time_t) -1
-               ? NONEXISTENT_MTIME
-               : file_timestamp_cons (file->hname, member_date, 0));
+
+      if (member_date == (time_t) -1
+          || (memmtime != NONEXISTENT_MTIME
+              && (time_t) FILE_TIMESTAMP_S (memmtime) > member_date))
+        /* If the member file exists and is newer than the member in the
+           archive, pretend it's nonexistent.  This means the member file was
+           updated but not added to the archive yet.  */
+        mtime = NONEXISTENT_MTIME;
+      else
+        mtime = file_timestamp_cons (file->hname, member_date, 0);
     }
   else
 #endif
@@ -1460,14 +1477,6 @@ f_mtime (struct file *file, int search)
       FILE_TIMESTAMP adjustment = FAT_ADJ_OFFSET << FILE_TIMESTAMP_LO_BITS;
       if (ORDINARY_MTIME_MIN + adjustment <= adjusted_mtime)
         adjusted_mtime -= adjustment;
-#elif defined(__EMX__)
-      /* FAT filesystems round time to the nearest even second!
-         Allow for any file (NTFS or FAT) to perhaps suffer from this
-         brain damage.  */
-      FILE_TIMESTAMP adjustment = (((FILE_TIMESTAMP_S (adjusted_mtime) & 1) == 0
-                     && FILE_TIMESTAMP_NS (adjusted_mtime) == 0)
-                    ? (FILE_TIMESTAMP) 1 << FILE_TIMESTAMP_LO_BITS
-                    : 0);
 #endif
 
       /* If the file's time appears to be in the future, update our
@@ -1546,7 +1555,7 @@ name_mtime (const char *name)
 
 #if defined(WINDOWS32)
   {
-    char tem[MAXPATHLEN], *tstart, *tend;
+    char tem[MAX_PATH+1], *tstart, *tend;
     const char *p = name + strlen (name);
 
     /* Remove any trailing slashes and "."/"..".  MS-Windows stat
