@@ -11,7 +11,7 @@
 #                         [-make <make prog>]
 #                        (and others)
 
-# Copyright (C) 1992-2022 Free Software Foundation, Inc.
+# Copyright (C) 1992-2023 Free Software Foundation, Inc.
 # This file is part of GNU Make.
 #
 # GNU Make is free software; you can redistribute it and/or modify it under
@@ -56,7 +56,7 @@ $memcheck_args = '--num-callers=15 --tool=memcheck --leak-check=full --suppressi
 $massif_args = '--num-callers=15 --tool=massif --alloc-fn=xmalloc --alloc-fn=xcalloc --alloc-fn=xrealloc --alloc-fn=xstrdup --alloc-fn=xstrndup';
 $pure_log = undef;
 
-# The location of the GNU make source directory
+# The location of the GNU Make source directory
 $srcdir = undef;
 $fqsrcdir = undef;
 $srcvol = undef;
@@ -104,79 +104,13 @@ if ($^O eq 'VMS')
 # We want them from the C locale regardless of our current locale.
 
 $ERR_no_such_file = undef;
+$ERR_no_such_file_code = "2";
 $ERR_read_only_file = undef;
 $ERR_unreadable_file = undef;
 $ERR_nonexe_file = undef;
 $ERR_exe_dir = undef;
 $ERR_command_not_found = undef;
 
-{
-  use locale;
-
-  my $loc = undef;
-  if ($has_POSIX) {
-      POSIX->import(qw(locale_h));
-      # Windows has POSIX locale, but only LC_ALL not LC_MESSAGES
-      $loc = POSIX::setlocale(&POSIX::LC_ALL);
-      POSIX::setlocale(&POSIX::LC_ALL, 'C');
-
-      # See set_defaults() as this doesn't work right on Windows :(
-      $! = &POSIX::ERANGE;
-  }
-
-  if (open(my $F, '<', 'file.none')) {
-      print "Opened non-existent file! Skipping related tests.\n";
-  } else {
-      $ERR_no_such_file = "$!";
-  }
-
-  unlink('file.out');
-  touch('file.out');
-
-  chmod(0444, 'file.out');
-  if (open(my $F, '>', 'file.out')) {
-      print "Opened read-only file! Skipping related tests.\n";
-      close($F);
-  } else {
-      $ERR_read_only_file = "$!";
-  }
-
-  $_ = `./file.out 2>&1`;
-  if ($? == 0) {
-      print "Executed non-executable file!  Skipping related tests.\n";
-  } else {
-      $ERR_nonexe_file = "$!";
-  }
-
-  $_ = `./. 2>&1`;
-  if ($? == 0) {
-      print "Executed directory!  Skipping related tests.\n";
-  } else {
-      $ERR_exe_dir = "$!";
-  }
-
-  chmod(0000, 'file.out');
-  if (open(my $F, '<', 'file.out')) {
-      print "Opened unreadable file!  Skipping related tests.\n";
-      close($F);
-  } else {
-      $ERR_unreadable_file = "$!";
-  }
-
-  unlink('file.out') or die "Failed to delete file.out: $!\n";
-
-  $_ = `/bin/sh -c 'bad-command 2>&1'`;
-  if ($? == 0) {
-      print "Invoked invalid file!  Skipping related tests.\n";
-  } else {
-      chomp($_);
-      s/bad-command/#CMDNAME#/g;
-      $ERR_command_not_found = $_;
-  }
-
-  $loc and POSIX::setlocale(&POSIX::LC_ALL, $loc);
-}
-
 #$SIG{INT} = sub { print STDERR "Caught a signal!\n"; die @_; };
 
 sub valid_option
@@ -195,7 +129,7 @@ sub valid_option
    if ($option =~ /^-srcdir$/i) {
        $srcdir = shift @argv;
        if (! -f File::Spec->catfile($srcdir, 'src', 'gnumake.h')) {
-           print "$option $srcdir: Not a valid GNU make source directory.\n";
+           print "$option $srcdir: Not a valid GNU Make source directory.\n";
            exit 1;
        }
        return 1;
@@ -252,7 +186,6 @@ sub subst_make_string
     s/#PERL#/$perl_name/g;
     s/#PWD#/$cwdpath/g;
     s/#WORK#/$workdir/g;
-    # If we're using a shell
     s/#HELPER#/$perl_name $helptool/g;
     return $_;
 }
@@ -472,6 +405,84 @@ sub set_defaults
   } else {
     $scriptsuffix = '.bat';
   }
+
+  $ENV{LC_ALL} = $makeENV{LC_ALL};
+  $ENV{LANG} = $makeENV{LANG};
+  $ENV{LANGUAGE} = $makeENV{LANGUAGE};
+
+  use locale;
+
+  my $loc = undef;
+  if ($has_POSIX) {
+      POSIX->import(qw(locale_h));
+      # Windows has POSIX locale, but only LC_ALL not LC_MESSAGES
+      $loc = POSIX::setlocale(&POSIX::LC_ALL);
+      POSIX::setlocale(&POSIX::LC_ALL, 'C');
+  }
+
+  if (open(my $F, '<', 'file.none')) {
+      print "Opened non-existent file! Skipping related tests.\n";
+  } else {
+      $ERR_no_such_file = "$!";
+      if ($osname eq 'os390') {
+          $ERR_no_such_file_code = "129";
+      }
+  }
+
+  unlink('file.out');
+  touch('file.out');
+
+  chmod(0444, 'file.out');
+  if (open(my $F, '>', 'file.out')) {
+      print "Opened read-only file! Skipping related tests.\n";
+      close($F);
+  } else {
+      $ERR_read_only_file = "$!";
+  }
+
+  $_ = `./file.out 2>&1`;
+  if ($? == 0) {
+      print "Executed non-executable file!  Skipping related tests.\n";
+  } else {
+      $ERR_nonexe_file = "$!";
+  }
+
+  if ($^O =~ /cygwin/i) {
+      # For some reason the execute here gives a different answer than make's
+      print "Skipping directory execution on $^O\n";
+  } else {
+      $_ = `./. 2>&1`;
+      if ($? == 0) {
+          print "Executed directory!  Skipping related tests.\n";
+      } else {
+          $ERR_exe_dir = "$!";
+      }
+  }
+
+  chmod(0000, 'file.out');
+  if (open(my $F, '<', 'file.out')) {
+      print "Opened unreadable file!  Skipping related tests.\n";
+      close($F);
+  } else {
+      $ERR_unreadable_file = "$!";
+  }
+
+  unlink('file.out') or die "Failed to delete file.out: $!\n";
+
+  $_ = `/bin/sh -c 'bad-command 2>&1'`;
+  if ($? == 0) {
+      print "Invoked invalid file!  Skipping related tests.\n";
+  } else {
+      s/\r?\n//g;
+      s/bad-command/#CMDNAME#/g;
+      $ERR_command_not_found = $_;
+  }
+
+  $loc and POSIX::setlocale(&POSIX::LC_ALL, $loc);
+
+  $ENV{LC_ALL} = $origENV{LC_ALL};
+  $ENV{LANG} = $origENV{LANG};
+  $ENV{LANGUAGE} = $origENV{LANGUAGE};
 }
 
 # This is no longer used: we import config-flags.pm instead
@@ -594,7 +605,7 @@ sub set_more_defaults
   }
 
   $string = `$make_path -v`;
-  $string =~ /^(GNU Make [^,\n]*)/ or die "$make_path is not GNU make.  Version:\n$string";
+  $string =~ /^(GNU Make [^,\n]*)/ or die "$make_path is not GNU Make.  Version:\n$string";
   $testee_version = "$1\n";
 
   create_file('null.mk', '');
